from typing import List
from collections import defaultdict

from .datastructures import \
    JunosSrxAddress, \
    JunosSrxAddressSet, \
    JunosSrxApplication, \
    JunosSrxRule

from .zone_providers import JunosSrxZoneProvider
from ...renderer import Renderer
from ....apps import Application
from ....rules import Rule


class JunosSrxRenderer(Renderer):
    META_CONTEXT = "jsrx_context"

    def __init__(self, **config):
        super().__init__(**config)
        self.addresses = {}
        self.address_sets = {}
        self.applications = {}

        self.global_policies = []
        self.zonal_policies = defaultdict(lambda: defaultdict(lambda: []))

        self.zone_provider = self._get_zone_provider()

    def _get_zone_provider(self) -> JunosSrxZoneProvider:
        return JunosSrxZoneProvider.get(
            self.config.get(
                "jsrx-zone-provider",
                "default"))

    def build_address_object(self, prefix: str) -> str:
        address = JunosSrxAddress.get_for_prefix(prefix, self.addresses)
        return address.name

    def build_address_set_object(
            self,
            set_name,
            address_names: List[str]) -> str:
        address_set = JunosSrxAddressSet(set_name, address_names)
        self.address_sets[set_name] = address_set
        return address_set.name

    def _initialise(self):
        self.preprocess_entities = True

    def _preprocess_prefix_list(self, pfx_list):
        names = [self.build_address_object(pfx) for pfx in pfx_list.prefixes]
        self.build_address_set_object(pfx_list.name, names)

    def _preprocess_app(self, app: Application):
        if "jsrx-app" not in app.metadata:
            junos_app = JunosSrxApplication.from_app(app)
            self.applications[app.name] = junos_app
            return app.name
        return app.metadata["jsrx-app"]

    def _process_rule(self, rule: Rule):
        apps = []

        for app in rule.applications.apps:
            if "jsrx-app" in app.metadata:
                apps.append(app.metadata["jsrx-app"])
            else:
                apps.append(self.applications[app.name].name)

        junos_rule = JunosSrxRule(
            rule.name,
            self.address_sets[rule.source.name].name,
            self.address_sets[rule.destination.name].name,
            apps,
            rule.action
        )

        if rule.metadata.get(self.META_CONTEXT, "zonal") == "global":
            junos_rule.force_global = True

        from_zones, to_zones = self.zone_provider.get_zones(rule)

        if from_zones:
            junos_rule.from_zones = from_zones

        if to_zones:
            junos_rule.to_zones = to_zones

        ctx = "global"
        seq = len(self.global_policies) + 1
        policy_list = self.global_policies

        if not junos_rule.is_global:
            policy_list = self.zonal_policies[from_zones[0]][to_zones[0]]
            ctx = f"{from_zones[0]}_{to_zones[0]}"
            seq = len(policy_list) + 1

        junos_rule.name = f"{ctx}_{seq}"

        policy_list.append(junos_rule)

    def get_sorted_objects(self):
        apps = list(self.applications.keys())
        addresses = list(self.addresses.keys())
        address_sets = list(self.address_sets.keys())

        apps.sort()
        addresses.sort()
        address_sets.sort()

        def _map(keys, dct):
            return map(lambda x: dct[x], keys)

        return (
            _map(apps, self.applications),
            _map(addresses, self.addresses),
            _map(address_sets, self.address_sets)
        )

    def get_sorted_zone_pairs(self):
        from_zones = list(self.zonal_policies.keys())
        from_zones.sort()

        for from_zone in from_zones:
            to_zones = list(self.zonal_policies[from_zone].keys())
            to_zones.sort()

            for to_zone in to_zones:
                yield (from_zone, to_zone)



from copy import deepcopy
from dataclasses import dataclass
from typing import Dict
from ..helpers import default
from ..apps import ApplicationLibrary
from ..prefixlists import PrefixListLoader
from ..rules import Rule, RuleSet
from ..renderers import Renderer
from ..artifacts import Artifact, ArtifactBlueprint
from ..constants import DEF_ENCODING


@dataclass
class Rulebook:
    title: str = "A Hyron Rulebook"
    owner: str = "Nobody"
    encoding: str = DEF_ENCODING
    apps: ApplicationLibrary = default(ApplicationLibrary)
    prefixlists: PrefixListLoader = default(PrefixListLoader)
    rules: Dict[str, Rule] = default(dict)
    rulesets: Dict[str, RuleSet] = default(dict)
    artifacts: Dict[str, ArtifactBlueprint] = default(dict)
    import_builtins: bool = True

    def build_artifact(self, name: str) -> Artifact:
        blueprint = self.artifacts[name]
        artifact = Artifact(self.encoding)
        artifact.meta = deepcopy(blueprint.meta)
        artifact.meta["_rulebook_title"] = self.title
        artifact.meta["_rulebook_owner"] = self.owner
        artifact.meta["_artifact_name"] = blueprint.name

        for name, filebp in blueprint.files.items():
            renderer = Renderer.get(filebp.renderer, **filebp.config)
            artifact_file = artifact[name]
            data = renderer.build(self.rulesets[filebp.ruleset])

            if renderer.BINARY:
                artifact_file.write_bytes(data)
            else:
                artifact_file.write_text(data)

        return artifact

    def build_all(self):
        return {name: self.build_artifact(name)
                for name in self.artifacts.keys()}



from py_mini_racer.py_mini_racer import MiniRacer

from .web import WebPrefixListDatasource


class JsonQueryListDatasource(WebPrefixListDatasource, register="jsonquery"):
    """
        This extension of the WebPrefixListDatasource allows users
        to embed Javascript in their prefix list definitions
        to filter and transform JSON data
    """
    def __init__(self, origin, query):
        self._ctx = self._build_v8_context(query)

        args = {
            "format": "json",
            "origin": origin
        }

        super().__init__(**args)

    @staticmethod
    def _build_v8_context(query) -> MiniRacer:
        js_code = f"let queryfunc = {query}"

        ctx = MiniRacer()
        ctx.eval(js_code)
        return ctx

    def _load_origin(self, origin):
        data = super()._load_origin(origin)
        result = self._ctx.call("queryfunc", data)
        assert(isinstance(result, list))
        return result



import os
from dataclasses import field
import radix

from ruamel.yaml import YAML


def _load(filename, handler):
    with open(filename, 'r') as infile:
        return handler(infile)


def _get_pkg_filename(*args):
    return os.path.join(get_pkg_dir(), *args)


def default(factory):
    return field(default_factory=factory)


def get_base_dir(filename):
    return os.path.dirname(os.path.realpath(filename))


def get_pkg_dir():
    return get_base_dir(__file__)


def optimise_prefixes(*prefix_list):
    rt = radix.Radix()

    for prefix in prefix_list:
        rt.add(prefix)

    return set([rt.search_worst(prefix).prefix for prefix in prefix_list])


def as_list(item):
    if isinstance(item, list):
        return item
    return [item]


def get_plural_dict_item(dic, name, single_handler=lambda x: [x]):
    plural = f"{name}s"
    if name in dic:
        return single_handler(dic[name])
    elif plural in dic:
        return dic[plural]
    raise KeyError(name)


def load_yaml(filename):
    yaml = YAML(typ="unsafe", pure=True)
    return _load(filename, lambda x: yaml.load(x))


def load_text(filename):
    return _load(filename, lambda x: x.read())


def get_builtin_filename(name):
    return _get_pkg_filename("builtin", f"{name}.yaml")


def get_asset_filename(name, ext="txt"):
    return _get_pkg_filename("assets", f"{name}.{ext}")


def on_dict_match(dic, key, value, if_retval, else_retval):
    if key in dic and dic[key] == value:
        return if_retval
    return else_retval


def resolve_working_directory(directory):
    if not directory:
        return os.getcwd()
    return directory



from .junos_srx_renderer import JunosSrxRenderer
from . import zone_providers, datastructures, formats

__all__ = [
    "JunosSrxRenderer",
    "zone_providers",
    "datastructures",
    "formats"
]




from .application import Application
from .application_group import ApplicationGroup


class ApplicationLibrary:
    def __init__(self):
        self._apps = {}
        self._appgroups = {}

    def register_app(self, name: str, app: Application):
        app.metadata["name"] = name
        self._apps[name] = app

    def register_appgroup(self, name: str, appgroup: ApplicationGroup):
        appgroup.metadata["name"] = name
        self._appgroups[name] = appgroup

    @property
    def apps(self):
        return set(self._apps.keys())

    @property
    def appgroups(self):
        return set(self._appgroups.keys())

    def __getitem__(self, item):
        if item in self._apps:
            return self._apps[item]
        elif item in self._appgroups:
            return self._appgroups[item]
        raise KeyError(item)

    def build_appgroup(self, name: str, members: list):
        apps = []
        for member in members:
            apps += self[member].apps
        appgroup = ApplicationGroup(apps)
        self.register_appgroup(name, appgroup)



from abc import ABC, abstractmethod
from typing import List
from ..constants import PROTOCOLS


class ApplicationContainer(ABC):
    @property
    @abstractmethod
    def apps(self) -> List["Application"]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def deterministic_name(self) -> str:
        raise NotImplementedError()

    def __hash__(self):
        return hash(self.deterministic_name)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.deterministic_name} at {hex(id(self))}>"  # noqa

    def __eq__(self, other):
        return self._unique_name == other._unique_name

    @property
    def _unique_name(self):
        if self.name != self.deterministic_name:
            return f"{self.name}({self.deterministic_name})"
        return self.deterministic_name


class Application(ApplicationContainer):
    ICMP = 1
    ICMPV6 = 53
    TUNIP4 = 4
    TCP = 6
    UDP = 17
    TUNIP6 = 41
    GRE = 47
    SCTP = 132

    def __init__(self, protocol: int):
        self._protocol = protocol
        self.metadata = {}

    def __le__(self, app):
        return self._protocol == app._protocol

    @property
    def apps(self):
        return [self]

    @property
    def name(self):
        if "name" in self.metadata:
            return self.metadata["name"]
        return self.deterministic_name

    @property
    def protocol(self):
        if self._protocol in PROTOCOLS:
            return PROTOCOLS[self._protocol].lower()
        return str(self._protocol)

    @property
    def protocol_id(self):
        return self._protocol

    @property
    def deterministic_name(self):
        return f"protocol_{self.protocol}"


class PortApplication(Application):
    def __init__(self, protocol: int, from_port: int, to_port: int):
        self.from_port = from_port
        self.to_port = to_port
        super().__init__(protocol)

    def __contains__(self, port):
        return self.from_port <= port <= self.to_port

    def __le__(self, app):
        if isinstance(app, type(self)):
            return self.protocol == app.protocol and self.to_port in app and self.from_port in app  # noqa
        return False

    @classmethod
    def tcp_port(cls, port):
        return cls(cls.TCP, port, port)

    @classmethod
    def tcp_port_range(cls, from_port, to_port):
        return cls(cls.TCP, from_port, to_port)

    @classmethod
    def udp_port(cls, port):
        return cls(cls.UDP, port, port)

    @classmethod
    def udp_port_range(cls, from_port, to_port):
        return cls(cls.UDP, from_port, to_port)

    @property
    def _name_suffix(self):
        if self.from_port == self.to_port:
            return f"port_{self.from_port}"
        return f"ports_{self.from_port}_to_{self.to_port}"

    @property
    def name(self):
        if "name" in self.metadata:
            return self.metadata["name"]
        return super().name

    @property
    def deterministic_name(self):
        return f"{super().deterministic_name}_{self._name_suffix}"  # noqa



from abc import abstractmethod
from typing import Tuple, List
from plugable import Plugable
from .....helpers import get_plural_dict_item, as_list
from .....rules import Rule


class JunosSrxZoneProvider(Plugable):
    META_FROM_ZONE = "jsrx_from_zones"
    META_TO_ZONE = "jsrx_to_zones"

    def get_zones(self, rule: Rule):
        from_zones, to_zones = self._get_zones(rule)

        if self.META_FROM_ZONE in rule.metadata:
            from_zones = as_list(rule.metadata[self.META_FROM_ZONE])

        if self.META_TO_ZONE in rule.metadata:
            to_zones = as_list(rule.metadata[self.META_TO_ZONE])

        return (from_zones, to_zones)

    @classmethod
    def _get_explicit_zones(cls, meta: dict) -> bool:
        from_zones = get_plural_dict_item(meta, cls.META_FROM_ZONE)
        to_zones = get_plural_dict_item(meta, cls.META_TO_ZONE)

        return (from_zones, to_zones)

    @abstractmethod
    def _get_zones(self, rule) -> Tuple[List[str]]:
        pass


class DefaultJunosSrxZoneProvider(JunosSrxZoneProvider, register="default"):
    def _get_zones(self, rule):
        return ([], [])



from .web import WebPrefixListDatasource
from ...constants import AWS_IP_RANGES


class AwsPrefixListDatasource(WebPrefixListDatasource, register="aws"):
    """
        This extension of the WebPrefixListDatasource allows users
        to easily reference and filter prefixes related to the
        various services and regions of Amazon Web Services
    """
    def __init__(self, region=None, service=None):
        self._region = region
        self._service = service

        args = {
            "format": "json",
            "origin": AWS_IP_RANGES
        }

        super().__init__(**args)

    def _load_origin(self, origin):
        data = super()._load_origin(origin)

        prefixes = []

        prefixes += [pfx["ip_prefix"]
                     for pfx in data["prefixes"] if self._is_valid(pfx)]
        prefixes += [pfx["ipv6_prefix"]
                     for pfx in data["ipv6_prefixes"] if self._is_valid(pfx)]

        return prefixes

    def _is_valid(self, prefix):
        if self._region and prefix["region"] != self._region:
            return False
        if self._service and prefix["service"] != self._service:
            return False
        return True



from .application import Application, PortApplication
from .application_library import ApplicationLibrary


_FROMPORT = "from-port"
_TOPORT = "to-port"
_PORT = "port"
_PORT_PROTOCOLS = {
    Application.TCP,
    Application.UDP
}


class ApplicationLibraryLoader:
    @staticmethod
    def _resolve_protocol(appdef):
        protocol = appdef["protocol"]
        if isinstance(protocol, int):
            return protocol
        else:
            protocol = protocol.upper()
            if hasattr(Application, protocol):
                return getattr(Application, protocol)
            else:
                raise NotImplementedError(f"{protocol} is not implemented")

    @staticmethod
    def _resolve_ports(appdef):
        if _FROMPORT in appdef and _TOPORT in appdef:
            return (appdef[_FROMPORT], appdef[_TOPORT])
        return (appdef[_PORT], appdef[_PORT])

    def _load_apps(self, appdefs: dict, lib):
        for name, appdef in appdefs.items():
            protocol = self._resolve_protocol(appdef)

            if protocol in _PORT_PROTOCOLS:
                ports = self._resolve_ports(appdef)
                app = PortApplication(protocol, *ports)
            else:
                app = Application(protocol)

            app.metadata.update(appdef.get("meta", {}))
            lib.register_app(name, app)

    def _load_appgroups(self, appgroupdefs: dict, lib):
        for name, applist in appgroupdefs.items():
            lib.build_appgroup(name, applist)

    def load(self, config: dict, lib: ApplicationLibrary = None):
        if not lib:
            lib = ApplicationLibrary()
        self._load_apps(config["apps"], lib)
        self._load_appgroups(config["appgroups"], lib)
        return lib



from .aws import AwsPrefixListDatasource
from .jsonquery import JsonQueryListDatasource
from .merge import MergePrefixListDatasource
from .static import StaticPrefixListDatasource
from .web import WebPrefixListDatasource

__all__ = [
    "AwsPrefixListDatasource",
    "JsonQueryListDatasource",
    "MergePrefixListDatasource",
    "StaticPrefixListDatasource",
    "WebPrefixListDatasource"
]



import os
import shutil
import tempfile
from testconstants import TEST_ARTIFACT_CONTENT, TEST_ARTIFACT_MANIFEST_CONTENT
from hyron.artifacts import Artifact
from hyron.constants import DEF_ENCODING


def _get_test_artifact():
    artifact = Artifact(DEF_ENCODING)
    newfile = artifact["test.json"]
    newfile.write_text(TEST_ARTIFACT_CONTENT)
    artifact.close_all()
    return artifact


def _assert_file_structure(directory):
    def get_absolute_fn(filename):
        return os.path.join(directory, filename)

    def assert_file(filename):
        assert(os.path.exists(get_absolute_fn(filename)))

    assert_file("meta.json")
    assert_file("files.json")
    assert_file("files/test.json")


def test_artifact_manifest():
    artifact = _get_test_artifact()
    manifest = artifact.to_manifest()
    assert(TEST_ARTIFACT_MANIFEST_CONTENT == manifest["files"]["test.json"])
    new_artifact = Artifact.from_manifest(manifest)
    file_data = new_artifact.files["test.json"].get_contents()
    assert(file_data.decode(DEF_ENCODING) == TEST_ARTIFACT_CONTENT)


def test_artifact_file_structure():
    artifact = _get_test_artifact()

    with tempfile.TemporaryDirectory() as tmpdir:
        artifact.to_file_structure(tmpdir)
        _assert_file_structure(tmpdir)


def test_artifact_archive():
    artifact = _get_test_artifact()

    with tempfile.TemporaryDirectory() as working_tmpdir:
        archive_file = artifact.to_archive(working_tmpdir)
        with tempfile.TemporaryDirectory() as target_tmpdir:
            shutil.unpack_archive(archive_file, target_tmpdir, "gztar")
            with os.scandir(target_tmpdir) as entries:
                for entry in entries:
                    print(entry.name)
            _assert_file_structure(target_tmpdir)



from typing import List
from ..constants import ACTION_PERMIT
from ..prefixlists.prefix_list import PrefixList
from ..apps.application import ApplicationContainer


class Rule:
    def __init__(
            self,
            src: PrefixList,
            dst: PrefixList,
            apps: ApplicationContainer,
            action=ACTION_PERMIT,
            **meta):
        self.source = src
        self.destination = dst
        self.applications = apps
        self.action = action
        self.metadata = meta

    @property
    def prefix_lists(self) -> List[PrefixList]:
        return [self.source, self.destination]

    @property
    def name(self) -> str:
        if "name" in self.metadata:
            return self.metadata["name"]
        app = self.applications.name
        src = self.source.name
        dst = self.destination.name
        return f"{self.action}_{app}_from_{src}_to_{dst}"



from typing import Any, List, Protocol, Text


class JunosSrxDatastructure(Protocol):
    def get_json_config_element(self) -> Any:
        raise NotImplementedError

    def get_set_commands(self) -> List[Text]:
        raise NotImplementedError



from io import BytesIO
from typing import BinaryIO


class ArtifactFile:
    def __init__(self, name, encoding):
        self.name = name
        self.encoding = encoding
        self.buf = BytesIO()
        self._open = True

    def _write(self, callback):
        if self._open:
            self.buf.write(callback())
        else:
            raise RuntimeError(f"File '{self.name}' is already closed")

    def write_bytes(self, data: bytes):
        self._write(lambda: data)

    def write_text(self, data: str):
        self._write(lambda: data.encode(self.encoding))

    def close(self):
        self._open = False

    def get_contents(self):
        self.close()
        return self.buf.getvalue()

    def dump(self, fh: BinaryIO):
        fh.write(self.get_contents())

    def load(self, fh: BinaryIO):
        self.write_bytes(fh.read())
        self.close()



from functools import reduce
from typing import Iterable

from ..junos_srx_renderer import JunosSrxRenderer
from ..datastructures import JunosSrxDatastructure


class JunosSrxCommandRenderer(JunosSrxRenderer, register="jsrx-cmd"):
    @staticmethod
    def flatten(lists):
        return reduce(lambda x, y: x + y, lists, [])

    @classmethod
    def build_commands(cls, datastructures: Iterable[JunosSrxDatastructure], apply_group=None):  # noqa
        return cls.flatten(
            map(lambda x: x.get_set_commands(apply_group), datastructures)
        )

    def _build_artifact(self):
        commands = []
        apply_group = None

        if "apply-group" in self.config:
            apply_group = str(self.config["apply-group"])
            commands.append(f"delete groups {apply_group}")
        else:
            commands = [
                "delete security policies",
                "delete security address-book global",
                "delete applications"
            ]

        apps, addresses, address_sets = self.get_sorted_objects()

        commands += self.build_commands(addresses, apply_group)
        commands += self.build_commands(address_sets, apply_group)
        commands += self.build_commands(apps, apply_group)
        commands += self.build_commands(self.global_policies, apply_group)

        for from_zone, to_zone in self.get_sorted_zone_pairs():
            policies = self.zonal_policies[from_zone][to_zone]
            commands += self.build_commands(policies, apply_group)

        if apply_group:
            commands.append(f"set apply-groups {apply_group}")

        return "\n".join(commands)



import pytest
import hyron
from testconstants import TEST_NONEXIST_APP


def _get_builtin_app_config():
    return hyron.helpers.load_yaml(
        hyron.helpers.get_builtin_filename("apps"))


def _get_app_loader():
    config = _get_builtin_app_config()["objects"]["appdefs"]
    return hyron.apps.ApplicationLibraryLoader().load(config)


def _get_ssh_app():
    return hyron.apps.PortApplication.tcp_port(22)


def _get_dns_app():
    return hyron.apps.PortApplication.tcp_port(53)


def _get_icmp4_app():
    return hyron.apps.Application(1)


def _get_anytcp_app():
    return hyron.apps.PortApplication.tcp_port_range(1, 65535)


def _get_anyudp_app():
    return hyron.apps.PortApplication.tcp_port_range(1, 65535)


def _get_demon_app():
    return hyron.apps.Application(666)


def test_builtin_apps_set_access():
    ldr = _get_app_loader()

    assert("any" in ldr.appgroups)
    assert("any-tcp" in ldr.apps)
    assert("any-udp" in ldr.apps)


def test_builtin_apps_nonexist():
    ldr = _get_app_loader()

    with pytest.raises(KeyError):
        ldr[TEST_NONEXIST_APP]


def test_builtin_names():
    ldr = _get_app_loader()

    names = list(ldr.apps) + list(ldr.appgroups)

    for name in names:
        assert(ldr[name].name == name)


def test_deterministic_names():
    answers = [
        (_get_ssh_app(), "protocol_tcp_port_22"),
        (_get_icmp4_app(), "protocol_icmp"),
        (_get_anytcp_app(), "protocol_tcp_ports_1_to_65535"),
        (_get_demon_app(), "protocol_666")
    ]

    names = []
    apps = []

    for obj, answer in answers:
        apps.append(obj)
        names.append(answer)
        assert(obj.name == answer)
        assert(hash(obj) == hash(answer))

    group = hyron.apps.ApplicationGroup(apps)
    names = ','.join(names)

    assert(group.name == f"group({names})")


def test_comparisons():
    assert(_get_icmp4_app() == _get_icmp4_app())
    assert(53 in _get_dns_app())
    assert(_get_ssh_app() <= _get_anytcp_app())



from dataclasses import dataclass
from typing import List

from ...junos_command_builder import JunosCommandBuilder
from .....helpers import default


@dataclass
class JunosSrxAddressSet:
    name: str
    address_names: List[str] = default(list)

    def get_json_config_element(self):
        return {
            "name": self.name,
            "address": [{"name": name} for name in self.address_names]
        }

    def get_set_commands(self, apply_group=None) -> List[str]:
        cmds = JunosCommandBuilder(apply_group)

        config_root = f"security address-book global address-set {self.name}"

        for address_name in self.address_names:
            cmds.add(f"{config_root} address {address_name}")

        return cmds.items



from metaloader.stanzahandlers.dict import DictStanzaHandler


class ObjectsStanzaHandler(DictStanzaHandler, register="objects"):
    def _initialise(self):
        return {
            "prefixlists": {},
            "appdefs": {
                "apps": {},
                "appgroups": {}
            }
        }

    def _merge(self, stanza_data: dict, new_data: dict):
        if "prefixlists" in new_data:
            stanza_data["prefixlists"].update(new_data["prefixlists"])
        if "appdefs" in new_data:
            appdefs = new_data["appdefs"]
            for key in ("apps", "appgroups"):
                if key in appdefs:
                    stanza_data["appdefs"][key].update(appdefs[key])



from .objects_stanza_handler import ObjectsStanzaHandler  # noqa

PROTOCOLS = {
    1: 'ICMP',
    4: 'TUNIP4',
    6: 'TCP',
    17: 'UDP',
    41: 'TUNIP6',
    47: 'GRE',
    132: 'SCTP'
}

AWS_IP_RANGES = "https://ip-ranges.amazonaws.com/ip-ranges.json"

ACTION_PERMIT = "permit"
ACTION_REJECT = "reject"
ACTION_IGNORE = "ignore"

ACTIONS = [
    ACTION_PERMIT,
    ACTION_REJECT,
    ACTION_IGNORE
]

DICT = "dict"
XDICT = {
    "name": DICT,
    "exclusive": True,
}

LOADER_SCHEMA = {
    "meta": XDICT,
    "artifacts": XDICT,
    "rules": DICT,
    "rulesets": DICT,
    "objects": "objects",
}

DEF_ENCODING = "utf8"



import hyron
import testconstants


def _get_builtin_prefixlist_config():
    return hyron.helpers.load_yaml(
        hyron.helpers.get_builtin_filename("prefixlists"))


def _get_prefixlist_loader():
    prefix_list_config = _get_builtin_prefixlist_config()
    prefix_list_config["objects"]["prefixlists"]["dummy"] = {
        "type": "__nonexist__"
    }
    return hyron.prefixlists.PrefixListLoader(
        prefix_list_config["objects"]["prefixlists"])


def test_prefixlist_providers():
    prefixlists = _get_prefixlist_loader()
    expected = len(_get_builtin_prefixlist_config()["objects"]["prefixlists"])

    assert(len(prefixlists.available) == expected)
    assert(len(prefixlists.loaded) == 0)

    for name, assertions in testconstants.PREFIX_LIST_TESTS.items():
        prefixlist = prefixlists[name]

        if "contains" in assertions:
            for prefix in assertions["contains"]:
                assert(prefix in prefixlist)
        if "not" in assertions:
            for prefix in assertions["not"]:
                assert(prefix not in prefixlist)

    # Merge providers will trigger loading of
    # other providers, therefore >= is used here
    assert(len(prefixlists.loaded) >= len(testconstants.PREFIX_LIST_TESTS))


def test_prefixlist():
    prefixlists = _get_prefixlist_loader()

    private = prefixlists["private"]

    assert(len(private.ipv4_prefixes) == 3)
    assert(len(private.ipv6_prefixes) == 1)

    assert("10.1.2.3" in private)
    assert("1.1.1.1" not in private)



from copy import copy

from .prefix_list import PrefixList
from .prefix_list_datasource import PrefixListDatasource


class PrefixListLoader:
    """
        Represents a collection of defined prefix lists.
        Lazily loads data as required.
    """

    def __init__(self, config: dict = None):
        self._datasources = {}
        self._cached = {}
        self._types = PrefixListDatasource.registry.enum()
        if config:
            self.include(config)

    def include(self, config: dict):
        for name, ds_config in config.items():
            ds_type = ds_config["type"]
            if ds_type in self._types:
                kwargs = copy(ds_config)
                del kwargs["type"]
                if ds_type == "merge":
                    # merge pfx lists require access to the loader to reference
                    # other prefix lists
                    kwargs["loader"] = self
                self._datasources[name] = PrefixListDatasource.get(
                    ds_type, **kwargs)

    @property
    def available(self):
        return set(self._datasources.keys())

    @property
    def loaded(self):
        return set(self._cached.keys())

    def __getitem__(self, key) -> PrefixList:
        if key not in self._cached:
            prefixes = self._datasources[key].fetch()
            self._cached[key] = PrefixList(prefixes, name=key)
        return self._cached[key]



from metaloader import FlatLoader, YamlSerialisation, LoaderContext
from .rule_builder import RuleBuilder, RuleSetBuilder
from .rulebook import Rulebook
from ..apps import ApplicationLibraryLoader
from ..artifacts import ArtifactBlueprint
from ..helpers import get_builtin_filename, load_yaml
from ..constants import LOADER_SCHEMA


class RulebookLoader:
    def __init__(self, fs=None):
        self._cfgldr = FlatLoader(YamlSerialisation())
        self._cfgldr.set_schema(LOADER_SCHEMA)
        self._appldr = ApplicationLibraryLoader()
        self._builtins = {}
        self.last_ctx: LoaderContext = None
        self.fs = fs

        self._load_builtin("apps")
        self._load_builtin("prefixlists")

    def _load_builtin(self, name):
        filename = get_builtin_filename(name)
        self._builtins[name] = load_yaml(filename)

    @staticmethod
    def _load_metadata(metadata: dict, rulebook: Rulebook):
        for key in ("title", "encoding", "import_builtins", "owner"):
            if key in metadata:
                setattr(rulebook, key, metadata[key])

    @staticmethod
    def _include_pfxlst_config(pfxlsts, rulebook: Rulebook):
        rulebook.prefixlists.include(pfxlsts)

    @staticmethod
    def _prepare_dict(objects, keys):
        for key in keys:
            if key not in objects:
                objects[key] = {}

    def _include_app_config(self, apps, rulebook: Rulebook):
        self._appldr.load(apps, rulebook.apps)

    def _get_include_handler(self, object_type):
        return {
            "appdefs": self._include_app_config,
            "prefixlists": self._include_pfxlst_config
        }[object_type]

    def _load_objects(self, objects: dict, rulebook: Rulebook):
        def apps(dct):
            return dct["appdefs"]["apps"]

        def appgroups(dct):
            return dct["appdefs"]["appgroups"]

        self._prepare_dict(objects, ["appdefs", "prefixlists"])
        self._prepare_dict(objects["appdefs"], ["apps", "appgroups"])

        if rulebook.import_builtins:
            apps(objects).update(apps(self._builtins["apps"]["objects"]))
            appgroups(objects).update(
                appgroups(self._builtins["apps"]["objects"]))
            objects["prefixlists"].update(
                self._builtins["prefixlists"]["objects"]["prefixlists"])

        for name, config in objects.items():
            self._get_include_handler(name)(config, rulebook)

    @staticmethod
    def _load_rules(rules: dict, rulebook: Rulebook):
        builder = RuleBuilder()
        for name, rule in rules.items():
            builder.add(name, rule)
        builder.resolve(rulebook)

    @staticmethod
    def _load_rulesets(rulesets: dict, rulebook: Rulebook):
        builder = RuleSetBuilder()
        for name, ruleset in rulesets.items():
            builder.add(name, ruleset)
        builder.resolve(rulebook)

    @staticmethod
    def _load_artifacts(artifacts: dict, rulebook: Rulebook):
        for name, config in artifacts.items():
            rulebook.artifacts[name] = ArtifactBlueprint.create(name, config)

    def load(self, filename: str) -> Rulebook:
        rulebook = Rulebook()
        rulebook.title = filename

        ctx = self._cfgldr.load(filename, self.fs)

        self._load_metadata(ctx.data["meta"], rulebook)
        self._load_objects(ctx.data["objects"], rulebook)
        self._load_rules(ctx.data["rules"], rulebook)
        self._load_rulesets(ctx.data["rulesets"], rulebook)
        self._load_artifacts(ctx.data["artifacts"], rulebook)

        self.last_ctx = ctx

        return rulebook



from ipaddress import ip_network
from dataclasses import dataclass
from typing import List

from ...junos_command_builder import JunosCommandBuilder


@dataclass
class JunosSrxAddress:
    name: str
    address: str

    @classmethod
    def get_for_prefix(cls, prefix: str, existing: dict):
        net = ip_network(prefix)
        network_addr = net.network_address.compressed
        name = f"pfx{net.version}-{network_addr}-{net.prefixlen}"

        if name not in existing:
            existing[name] = cls(name, net.exploded)

        return existing[name]

    def get_json_config_element(self):
        return {
            "name": self.name,
            "ip-prefix": self.address
        }

    def get_set_commands(self, apply_group=None) -> List[str]:
        cmds = JunosCommandBuilder(apply_group)

        cmds.add(f"security address-book global address {self.name} {self.address}")  # noqa

        return cmds.items



import os
import json

TEST_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

PREFIX_LIST_TESTS = {
    "cloudflare": {
        "contains": ["104.16.0.0/32", "2400:cb00::/128"]
    },
    "fullbogons": {
        "contains": ["10.20.30.40/32", "::1/128"],
        "not": ["8.8.8.8/32"]
    },
    "any": {
        "contains": ["1.1.1.1/32", "ff80::1111/128"]
    },
    "aws": {
        "contains": ["18.232.234.12/32"]
    },
    "aws-sydney-ec2": {
        "contains": ["13.236.45.1/32"],
        "not": ["18.232.234.12/32"]
    }
}

TEST_ARTIFACT_CONTENT = json.dumps({
    "hello": "world",
    "foo": {
        "bar": ["baz"]
    }
})

TEST_ARTIFACT_MANIFEST_CONTENT = "eyJoZWxsbyI6ICJ3b3JsZCIsICJmb28iOiB7ImJhciI6IFsiYmF6Il19fQ=="  # noqa

TEST_NONEXIST_APP = "__this_app_shouldn't_exist__"



from .prefix_list import PrefixList
from .prefix_list_loader import PrefixListLoader
from .prefix_list_datasource import PrefixListDatasource
from . import datasources


__all__ = [
    "PrefixList",
    "PrefixListLoader",
    "PrefixListDatasource",
    "datasources"
]



import json
from ..junos_srx_renderer import JunosSrxRenderer


class JunosSrxJsonRenderer(JunosSrxRenderer, register="jsrx"):
    @staticmethod
    def build_json_data(datastructures):
        return [x.get_json_config_element() for x in datastructures]

    def _build_artifact(self):
        zonal_policy_objects = []

        for from_zone, to_zone in self.get_sorted_zone_pairs():
            policies = self.zonal_policies[from_zone][to_zone]
            zonal_policy_objects.append({
                "from-zone-name": from_zone,
                "to-zone-name": to_zone,
                "policy": self.build_json_data(policies)
            })

        apps, addresses, address_sets = self.get_sorted_objects()

        artifact = {
            "applications": {
                "application": self.build_json_data(apps)
            },
            "security": {
                "address-book": [
                    {
                        "name": "global",
                        "address": self.build_json_data(addresses),
                        "address-set": self.build_json_data(address_sets)
                    }
                ],
                "policies": {
                    "policy": zonal_policy_objects
                }
            }
        }

        if self.global_policies:
            artifact["security"]["policies"]["global"] = {
                "policy": self.build_json_data(self.global_policies)
            }

        # Workaround bug in Junos for handling empty lists in JSON format
        if not artifact["applications"]["application"]:
            del artifact["applications"]

        if not artifact["security"]["policies"]["policy"]:
            del artifact["security"]["policies"]["policy"]

        if "apply-group" in self.config:
            applygrp = {"name": str(self.config["apply-group"])}
            applygrp.update(artifact)
            artifact = {"groups": [applygrp]}

        return json.dumps({"configuration": artifact}, indent=4)



from typing import Dict
from dataclasses import dataclass
from .artifact_file_blueprint import ArtifactFileBlueprint


@dataclass
class ArtifactBlueprint:
    name: str
    meta: Dict[str, str]
    files: Dict[str, ArtifactFileBlueprint]

    @classmethod
    def create(cls, name, config):
        files = {k: ArtifactFileBlueprint(k, **v)
                 for k, v in config["files"].items()}
        return cls(name, config["meta"], files)



from typing import List
from .rule import Rule
from ..prefixlists.prefix_list import PrefixList


class RuleSet:
    def __init__(self, rules=[], **meta):
        self.rules: List[Rule] = rules
        self.metadata = meta

    @property
    def prefix_lists(self) -> List[PrefixList]:
        pfx_lists = []
        for lists in map(lambda x: x.prefix_lists, self.rules):
            for pfx_list in lists:
                if pfx_list not in pfx_lists:
                    pfx_lists.append(pfx_list)
        return pfx_lists

    @property
    def applications(self):
        apps = []
        for appcontainer in map(lambda x: x.applications, self.rules):
            for app in appcontainer.apps:
                if app not in apps:
                    apps.append(app)
        return apps



from .application import Application, PortApplication, ApplicationContainer  # noqa
from .application_group import ApplicationGroup
from .application_library import ApplicationLibrary
from .application_library_loader import ApplicationLibraryLoader  # noqa

__all__ = [
    "Application",
    "ApplicationContainer",
    "ApplicationGroup",
    "ApplicationLibrary",
    "PortApplication",
    "ApplicationLibraryLoader"
]



import json
from ipaddress import ip_network

from ..renderer import Renderer
from ...rules.rule_set import Rule
from ...prefixlists.prefix_list import PrefixList
from ...apps.application import Application, PortApplication
from ...constants import ACTION_PERMIT

__all__ = [
    "AwsEc2SecurityGroupRenderer"
]


def _handle_app(app: Application):
    return {
        "IpProtocol": str(app.protocol)
    }


def _handle_port_app(app: PortApplication):
    args = _handle_app(app)
    args.update({
        "FromPort": app.from_port,
        "ToPort": app.to_port
    })
    return args


_APP_HANDLERS = {
    Application.TCP: _handle_port_app,
    Application.UDP: _handle_port_app
}

_CIDR_NAMES = {
    4: "CidrIp",
    6: "CidrIpv6"
}

_LIST_NAMES = {
    4: "IpRanges",
    6: "Ipv6Ranges"
}


def _build_ranges(desc, pfxlist: PrefixList):
    ranges = {}
    for prefix in pfxlist.prefixes:
        ver = ip_network(prefix).version
        range_type = _LIST_NAMES[ver]

        if range_type not in ranges:
            ranges[range_type] = []

        ranges[range_type].append({
            _CIDR_NAMES[ver]: prefix,
            "Description": desc
        })
    return ranges


class AwsEc2SecurityGroupRenderer(Renderer, register="ec2sg"):
    NAMED_IP_PROTOCOLS = {
        "icmp": 1,
        "icmpv6": 53,
        "tcp": 6,
        "udp": 17
    }

    def __init__(self, **config):
        super().__init__(**config)
        self.egress = False
        self.preprocess_entities = True
        self.description = ""
        self.application_entries = {}

    def _initialise(self):
        self.permissions = []
        self.egress = self.config.get("direction", "ingress") == "egress"
        self.description = self.metadata.get(
            "description", self.config.get(
                "description", "hyron"))

    def _preprocess_prefix_list(self, prefixes):
        pass

    def _preprocess_app(self, app: Application):
        entry = {
            "IpRanges": [],
            "Ipv6Ranges": [],
            "PrefixListIds": [],
            "UserIdGroupPairs": [],
        }
        entry.update(_APP_HANDLERS.get(app.protocol_id, _handle_app)(app))
        self.application_entries[app.name] = entry

    def _process_rule(self, rule: Rule):
        if rule.action == ACTION_PERMIT:
            for app in rule.applications.apps:
                interesting_side = "source"
                direction = "from"

                if self.egress:
                    interesting_side = "destination"
                    direction = "to"

                pfxs = getattr(rule, interesting_side)
                desc = f"{self.description}_{app.name}_{direction}_{pfxs.name}"

                permission = self.application_entries[app.name]
                ranges = _build_ranges(desc, pfxs)

                for range_type in ("IpRanges", "Ipv6Ranges"):
                    permission[range_type] += ranges.get(range_type, [])

    def _build_artifact(self):
        return json.dumps(list(self.application_entries.values()), indent=4)

    @classmethod
    def _format_protocol(cls, protocol):
        if protocol in cls.NAMED_IP_PROTOCOLS:
            return cls.NAMED_IP_PROTOCOLS[protocol]
        return int(protocol)



from .rulebook import Rulebook
from ..rules import Rule, RuleSet
from ..prefixlists import PrefixList
from ..constants import ACTION_PERMIT, ACTIONS

_DEF_PFX = PrefixList(["0.0.0.0/0", "::/0"], name="any")


class RuleBuilder:
    def __init__(self):
        self._proto_rules = {}

    def add(self, name: str, rule: dict):
        src = rule.get("src", None)
        dst = rule.get("dst", None)
        action = rule.get("action", ACTION_PERMIT)
        meta = rule.get("meta", {})
        app = rule["app"]

        assert(action in ACTIONS)

        self._proto_rules[name] = (src, dst, app, action, meta)

    def resolve(self, rulebook: Rulebook):
        rules = {}

        for name, (src, dst, app, action, meta) in self._proto_rules.items():
            src_pfxs = _DEF_PFX
            dst_pfxs = _DEF_PFX

            if src:
                src_pfxs = rulebook.prefixlists[src]

            if dst:
                dst_pfxs = rulebook.prefixlists[dst]

            rules[name] = Rule(
                src_pfxs,
                dst_pfxs,
                rulebook.apps[app],
                action,
                name=name,
                **meta)

        self._proto_rules = {}
        rulebook.rules.update(rules)


class RuleSetBuilder:
    def __init__(self):
        self._proto_rule_sets = {}

    def add(self, name: str, ruleset: dict):
        self._proto_rule_sets[name] = (
            ruleset["rules"], ruleset.get("meta", {}))

    def resolve(self, rulebook: Rulebook):
        rule_sets = {
            setname: RuleSet([rulebook.rules[name] for name in rulenames], **meta)  # noqa
            for setname, (rulenames, meta) in self._proto_rule_sets.items()
        }

        self._proto_rule_sets = {}
        rulebook.rulesets.update(rule_sets)



from .junos_srx_address import JunosSrxAddress
from .junos_srx_address_set import JunosSrxAddressSet
from .junos_srx_application import JunosSrxApplication
from .junos_srx_rule import JunosSrxRule
from .junos_srx_datastructure import JunosSrxDatastructure

__all__ = [
    "JunosSrxAddress",
    "JunosSrxAddressSet",
    "JunosSrxApplication",
    "JunosSrxRule",
    "JunosSrxDatastructure"
]



class JunosCommandBuilder:
    def __init__(self, apply_group=None):
        self.apply_group = apply_group
        self.items = []

    @property
    def set_prefix(self):
        if self.apply_group is not None:
            return f"set groups {self.apply_group}"
        return "set"

    def add(self, cmd, no_append=False, front=False):
        if not no_append:
            cmd = f"{self.set_prefix} {cmd}"

        if front:
            self.items.insert(0, cmd)
        else:
            self.items.append(cmd)



import os
import json
import shutil
import tempfile
from copy import deepcopy
from base64 import b64encode, b64decode
from typing import Dict
from .artifact_file import ArtifactFile
from ..constants import DEF_ENCODING
from ..helpers import resolve_working_directory


class Artifact:
    def __init__(self, encoding=DEF_ENCODING):
        self.encoding = encoding
        self.meta = {}
        self.files: Dict[str, ArtifactFile] = {}

    def __getitem__(self, key) -> ArtifactFile:
        if key not in self.files:
            self.files[key] = ArtifactFile(key, self.encoding)
        return self.files[key]

    def close_all(self):
        for file_obj in self.files.values():
            file_obj.close()

    def _get_encodable_meta(self):
        meta = deepcopy(self.meta)
        meta["_encoding"] = self.encoding
        return meta

    def to_manifest(self):
        def encode(file_obj: ArtifactFile):
            return b64encode(file_obj.get_contents()).decode("ascii")

        return {
            "meta": self._get_encodable_meta(),
            "files": {k: encode(v) for k, v in self.files.items()}
        }

    @property
    def name(self):
        return self.meta.get("_artifact_name", "artifact")

    @classmethod
    def from_manifest(cls, manifest: dict) -> "Artifact":
        def decode(data: str):
            return b64decode(data.encode("ascii"))

        artifact = cls(manifest["meta"]["_encoding"])

        for k, v in manifest["files"].items():
            file_obj = artifact[k]
            file_obj.write_bytes(decode(v))

        artifact.close_all()

        return artifact

    def to_json(self) -> str:
        return json.dumps(self.to_manifest())

    def to_file_structure(self, rootdir=None):
        rootdir = resolve_working_directory(rootdir)

        def get_path(filename):
            return os.path.join(rootdir, filename)

        def dump_file(filename, content, mode="w", handler=json.dump):
            absolute_fn = get_path(filename)
            with open(absolute_fn, mode) as outfile:
                handler(content, outfile)

        dump_file("meta.json", self._get_encodable_meta())
        dump_file("files.json", list(self.files.keys()))

        os.mkdir(get_path("files"))

        for k, v in self.files.items():
            dump_file(f"files/{k}", v, "wb", lambda x, y: x.dump(y))

    def to_archive(self, rootdir=None, compression="gztar"):
        filename = os.path.join(resolve_working_directory(rootdir), self.name)

        with tempfile.TemporaryDirectory() as tmpdir:
            self.to_file_structure(tmpdir)
            return shutil.make_archive(filename, compression, tmpdir)



from . import apps, prefixlists, rules, renderers, rulebooks, artifacts, constants, helpers  # noqa

__all__ = [
    "apps",
    "prefixlists",
    "rules",
    "renderers",
    "rulebooks",
    "artifacts",
    "constants",
    "helpers"
]



from dataclasses import dataclass
from typing import Optional, List

from ...junos_command_builder import JunosCommandBuilder
from .....apps import Application, PortApplication


@dataclass
class JunosSrxApplication:
    name: str
    protocol: str
    from_port: Optional[str] = None
    to_port: Optional[str] = None

    @classmethod
    def from_app(cls, app: Application):
        this = cls(app.name, app.protocol)

        if isinstance(app, PortApplication):
            this.from_port = app.from_port
            this.to_port = app.to_port

        return this

    @property
    def destination_port_content(self):
        if self.from_port:
            dstport = self.from_port
            if self.to_port and self.to_port != dstport:
                dstport = f"{dstport}-{self.to_port}"
            return dstport
        return None

    def get_json_config_element(self):
        element = {
            "name": self.name,
            "protocol": self.protocol,
        }

        dstport = self.destination_port_content

        if dstport:
            element["destination-port"] = dstport

        return element

    def get_set_commands(self, apply_group=None) -> List[str]:
        cmds = JunosCommandBuilder(apply_group)
        config_root = f"applications application {self.name}"
        dstport = self.destination_port_content

        cmds.add(f"{config_root} protocol {self.protocol}")
        if dstport:
            cmds.add(f"{config_root} destination-port {dstport}")

        return cmds.items



import json
import requests

from ..prefix_list_datasource import PrefixListDatasource
from ...helpers import get_plural_dict_item


class WebPrefixListDatasource(PrefixListDatasource, register="web"):
    """
        This datasource allows users to retrieve prefix list data from
        web locations
    """
    FORMAT_HANDLERS = {
        "text": lambda x: [
            i for i in map(
                lambda y: y.strip(),
                x.splitlines()) if i[0] != "#"],
        "csv": lambda x: x.strip().split(","),
        "json": lambda x: json.loads(x)}

    def __init__(self, **kwargs):
        self._origins = get_plural_dict_item(kwargs, "origin")
        self._format = kwargs["format"]
        self._session = requests.Session()

    def _load_origin(self, origin):
        response = self._session.get(origin)
        encoding = response.apparent_encoding
        raw = response.content.decode(encoding)
        return self.FORMAT_HANDLERS[self._format](raw)

    def fetch(self):
        prefixes = []
        for origin in self._origins:
            prefixes += self._load_origin(origin)
        return prefixes



import ipaddress
from collections import defaultdict

import radix

from ..helpers import optimise_prefixes


class PrefixList:
    NAME_PREFIX = "nets"

    def __init__(self, prefixes, **meta):
        self._prefixes = optimise_prefixes(*prefixes)
        self._ipv4_prefixes = []
        self._ipv6_prefixes = []
        self._reorder_prefixes()

        self._radix = radix.Radix()
        for prefix in self._prefixes:
            self._radix.add(prefix)
        self.metadata = meta
        self._repr = self._calculate_repr()

    @staticmethod
    def _explode(nets: list):
        return list(map(lambda x: x.exploded, nets))

    def _calculate_repr(self):
        joined_pfxs = ",".join(self.prefixes)
        return f"{self.name}({joined_pfxs})"

    def __repr__(self):
        return self._repr

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __contains__(self, prefix):
        if self.search(prefix):
            return True
        return False

    @property
    def prefixes(self):
        return self._prefixes

    def _reorder_prefixes(self):
        nets = defaultdict(lambda: [])

        for ip_network in map(ipaddress.ip_network, self._prefixes):
            nets[ip_network.version].append(ip_network)

        for net_list in nets.values():
            net_list.sort()

        self._ipv4_prefixes = self._explode(nets[4])
        self._ipv6_prefixes = self._explode(nets[6])

        self._prefixes = self._ipv4_prefixes + self._ipv6_prefixes

    @property
    def ipv4_prefixes(self):
        return self._ipv4_prefixes

    @property
    def ipv6_prefixes(self):
        return self._ipv6_prefixes

    def search(self, prefix):
        result = self._radix.search_best(prefix)
        if result:
            result = result.prefix
        return result

    @property
    def name(self):
        name = self.metadata["name"]
        return f"{self.NAME_PREFIX}_{name}"



from ..prefix_list_datasource import PrefixListDatasource


class StaticPrefixListDatasource(PrefixListDatasource, register="static"):
    """
        This datasource allows users to define static prefix lists
    """
    def __init__(self, items):
        self._items = items

    def fetch(self):
        return self._items



import os
import hyron
import metaloader
import testconstants


def _get_test_rulebook_path(name):
    return os.path.join(
        testconstants.TEST_DIRECTORY,
        "data",
        "rulebooks",
        name,
        "main.yaml")


def _load_rulebook(filename):
    loader = hyron.rulebooks.RulebookLoader()
    return loader.load(filename)


def _load_datastructure(filename):
    loader = metaloader.FlatLoader(
        serialisation=metaloader.YamlSerialisation())
    loader.set_schema(hyron.constants.LOADER_SCHEMA)
    return loader.load(filename).data


def test_integration():
    filename = _get_test_rulebook_path("simple")
    rulebook = _load_rulebook(filename)
    data = _load_datastructure(filename)
    assert(rulebook.title == data["meta"]["title"])
    assert(rulebook.owner == data["meta"]["owner"])
    artifacts = rulebook.build_all()

    for name in data["artifacts"].keys():
        assert(name in artifacts)



from functools import reduce

from ..prefix_list_loader import PrefixListLoader
from ..prefix_list_datasource import PrefixListDatasource


class MergePrefixListDatasource(PrefixListDatasource, register="merge"):
    """
        This datasource allows users to merge other datasources
    """
    def __init__(self, names, loader: PrefixListLoader):
        self._names = names
        self._loader = loader

    def fetch(self):
        prefix_lists = [self._loader[name] for name in self._names]
        return reduce(lambda x, y: x + list(y.prefixes), prefix_lists, [])



from abc import abstractmethod
from typing import Dict
from plugable import Plugable
from ..rules.rule_set import RuleSet, Rule
from ..apps.application import Application
from ..prefixlists.prefix_list import PrefixList
from ..helpers import on_dict_match


class Renderer(Plugable):
    """
        Renderers take compiled rule-set objects and output one
        "artifact", which is a text-based representation of rules in a
        format native to a device or system capable of enforcing them.
    """
    BINARY = False  # Set this flag to true if your Renderer outputs binary
    HINTS = {}

    def __init__(self, **config):
        self.config = config
        self.metadata = None
        self.preprocess_entities = False

    def build(self, rule_set: RuleSet) -> Dict[str, str]:
        self.metadata = rule_set.metadata

        self._initialise()

        if self.preprocess_entities:
            for pfx_list in rule_set.prefix_lists:
                self._preprocess_prefix_list(pfx_list)
            for app in rule_set.applications:
                self._preprocess_app(app)

        for rule in rule_set.rules:
            self._process_rule(rule)

        return self._build_artifact()

    def _assert_metadata(self, key, value, if_retval, else_retval):
        return on_dict_match(self.metadata, key, value, if_retval, else_retval)

    def _initialise(self):
        return

    def _preprocess_prefix_list(self, prefixes: PrefixList):
        raise NotImplementedError

    def _preprocess_app(self, app: Application):
        raise NotImplementedError

    @abstractmethod
    def _process_rule(self, rule: Rule):
        pass

    @abstractmethod
    def _build_artifact(self):
        pass



from .application import ApplicationContainer


class ApplicationGroup(ApplicationContainer):
    def __init__(self, apps=[]):
        self._apps = apps
        self.metadata = {}

    @property
    def apps(self):
        return self._apps

    @property
    def name(self):
        if "name" in self.metadata:
            return self.metadata["name"]
        return self.deterministic_name

    @property
    def deterministic_name(self):
        joined_apps = ",".join(map(lambda x: x.deterministic_name, self.apps))
        return f"group({joined_apps})"

    # TODO: Add application group optimisation logic, deterministic sorting



from dataclasses import dataclass
from typing import List, Optional, Text

from ...junos_command_builder import JunosCommandBuilder


@dataclass
class JunosSrxRule:
    description: str
    source: str
    destination: str
    applications: List[str]
    action: str
    force_global: bool = False
    from_zones: Optional[List[str]] = None
    to_zones: Optional[List[str]] = None
    comment: Optional[str] = None
    name: Optional[str] = None

    @property
    def is_global(self):
        if not self.force_global and self.from_zones and self.to_zones:
            return len(self.from_zones) != 1 or len(self.to_zones) != 1
        return True

    def get_json_config_element(self):
        element = {
            "name": self.name if self.name else self.description,
            "description": self.description,
            "match": {
                "source-address": [self.source],
                "destination-address": [self.destination],
                "application": self.applications
            },
            "then": {
                self.action: [None]
            }
        }

        if self.is_global:
            element["match"]["from-zones"] = self.from_zones
            element["match"]["to-zones"] = self.to_zones

        if self.comment:
            element["@"] = {
                "comment": self.comment
            }

        return element

    def get_set_commands(self, apply_group=None) -> List[str]:
        cmds = JunosCommandBuilder(apply_group)
        name = self.name if self.name else self.description
        config_root = "security policies"

        def render_global_zones(zones: Optional[List[Text]], verb: str):
            if zones:
                for zone in zones:
                    cmds.add(f"{config_root} match {verb}-zone {zone}")

        if self.is_global:
            config_root = f"{config_root} global policy {name}"
            render_global_zones(self.from_zones, "from")
            render_global_zones(self.to_zones, "to")
        else:
            from_zone = self.from_zones[0]
            to_zone = self.to_zones[0]
            config_root = f"{config_root} from-zone {from_zone} to-zone {to_zone} policy {name}"  # noqa

        cmds.add(f"{config_root} match source-address {self.source}")
        cmds.add(f"{config_root} match destination-address {self.destination}")

        for app in self.applications:
            cmds.add(f"{config_root} match application {app}")

        cmds.add(f"{config_root} then {self.action}")

        return cmds.items
